/********************************************************************************
*                                                                               *
*                          F O X   T e x t   E d i t o r                        *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998,2000 by Jeroen van der Zijp.   All Rights Reserved.        *
*********************************************************************************
* $Id: TextEdit.h,v 1.11 2000/10/28 06:27:28 jeroen Exp $                       *
********************************************************************************/
#ifndef TEXTEDIT_H
#define TEXTEDIT_H


/*******************************************************************************/

class HelpWindow;
  

// Mini application object
class FXAPI TextWindow : public FXMainWindow {
  FXDECLARE(TextWindow)
protected:
  HelpWindow          *helpwindow;          // Help window
  FXMenuPane          *filemenu;            // File menu
  FXMenuPane          *editmenu;            // Edit menu
  FXMenuPane          *optionmenu;          // Option menu
  FXMenuPane          *viewmenu;            // View menu
  FXMenuPane          *helpmenu;            // Help menu
  FXVerticalFrame     *treebox;             // Tree box containing directories/files
  FXText              *editor;              // Multiline text widget
  FXDirList           *dirlist;             // Directory view
  FXTextField         *filter;              // Filter for tree list
  FXIcon              *bigicon;             // Big application icon
  FXIcon              *smallicon;           // Small application icon
  FXIcon              *newicon;
  FXIcon              *openicon;
  FXIcon              *saveicon;
  FXIcon              *saveasicon;
  FXIcon              *printicon;
  FXIcon              *cuticon;
  FXIcon              *copyicon;
  FXIcon              *pasteicon;
  FXIcon              *deleteicon;
  FXIcon              *undoicon;
  FXIcon              *redoicon;
  FXIcon              *fontsicon;
  FXIcon              *colorsicon;
  FXIcon              *helpicon;
  FXUndoList           undolist;            // Undo list
  FXRecentFiles        mrufiles;            // Recent files
  FXString             filename;            // File being edited
  FXbool               namechanged;         // Save becomes save as if no name given yet
  FXbool               modified;            // Modified state
  FXbool               stripcr;             // Strip carriage returns
  static const FXchar *patterns[];          // File patterns
protected:
  FXbool loadFile(const FXString& file);
  FXbool saveFile(const FXString& file);
  void readRegistry();
  void writeRegistry();
protected:
  enum{
    MAXUNDOSIZE    = 1000000,               // Don't let the undo buffer get out of hand
    KEEPUNDOSIZE   = 500000                 // When MAXUNDOSIZE was exceeded, trim down to this size
    };
protected:
  TextWindow(){}
public:
  long onCmdAbout(FXObject*,FXSelector,void*);
  long onCmdRevert(FXObject*,FXSelector,void*);
  long onUpdRevert(FXObject*,FXSelector,void*);
  long onCmdOpen(FXObject*,FXSelector,void*);
  long onCmdSave(FXObject*,FXSelector,void*);
  long onUpdSave(FXObject*,FXSelector,void*);
  long onCmdSaveAs(FXObject*,FXSelector,void*);
  long onCmdNew(FXObject*,FXSelector,void*);
  long onCmdFont(FXObject*,FXSelector,void*);
  long onCmdQuit(FXObject*,FXSelector,void*);
  long onCmdPrint(FXObject*,FXSelector,void*);
  long onUpdTitle(FXObject*,FXSelector,void*);
  long onCmdTreeList(FXObject*,FXSelector,void*);
  long onCmdBackColor(FXObject*,FXSelector,void*);
  long onUpdBackColor(FXObject*,FXSelector,void*);
  long onCmdTextColor(FXObject*,FXSelector,void*);
  long onUpdTextColor(FXObject*,FXSelector,void*);
  long onCmdSelBackColor(FXObject*,FXSelector,void*);
  long onUpdSelBackColor(FXObject*,FXSelector,void*);
  long onCmdSelTextColor(FXObject*,FXSelector,void*);
  long onUpdSelTextColor(FXObject*,FXSelector,void*);
  long onCmdLineColor(FXObject*,FXSelector,void*);
  long onUpdLineColor(FXObject*,FXSelector,void*);
  long onCmdRecentFile(FXObject*,FXSelector,void*);
  long onCmdSaveSettings(FXObject*,FXSelector,void*);
  long onTextInserted(FXObject*,FXSelector,void*);
  long onTextDeleted(FXObject*,FXSelector,void*);
  long onCmdSearch(FXObject*,FXSelector,void*);
  long onCmdReplace(FXObject*,FXSelector,void*);
  long onCmdWrap(FXObject*,FXSelector,void*);
  long onUpdWrap(FXObject*,FXSelector,void*);
  long onCmdWrapFixed(FXObject*,FXSelector,void*);
  long onUpdWrapFixed(FXObject*,FXSelector,void*);
  long onEditDNDMotion(FXObject*,FXSelector,void*);
  long onEditDNDDrop(FXObject*,FXSelector,void*);
  long onCmdStripReturns(FXObject*,FXSelector,void*);
  long onUpdStripReturns(FXObject*,FXSelector,void*);
  
public:
  enum{
    ID_ABOUT=FXMainWindow::ID_LAST,
    ID_OPEN,
    ID_REVERT,
    ID_SAVE,
    ID_SAVEAS,
    ID_NEW,
    ID_TITLE,
    ID_FONT,
    ID_QUIT,
    ID_PRINT,
    ID_TREELIST,
    ID_BACKCOLOR,
    ID_TEXTCOLOR,
    ID_SELBACKCOLOR,
    ID_SELTEXTCOLOR,
    ID_LINECOLOR,
    ID_RECENTFILE,
    ID_TOGGLE_WRAP,
    ID_FIXED_WRAP,
    ID_SAVE_SETTINGS,
    ID_TEXT,
    ID_SEARCH,
    ID_REPLACE,
    ID_STRIP_CR,
    ID_LAST
    };
public:
  TextWindow(FXApp* a);
  virtual void create();
  void start(int& argc,char** argv);
  virtual ~TextWindow();
  };

#endif

